@@ -5,83 +5,13 @@ import warnings
 from redis import StrictRedis
 import numpy as np
 
-from . import utils
-from .command_builder import Builder
+from redisai import command_builder as builder
+from redisai.dag import Dag
+from redisai.pipeline import Pipeline
+from redisai.postprocessor import Processor
 
 
-builder = Builder()
-
-
-def enable_debug(f):
-    @wraps(f)
-    def wrapper(*args):
-        print(*args)
-        return f(*args)
-    return wrapper
-
-
-class Dag:
-    def __init__(self, load, persist, executor, readonly=False):
-        self.result_processors = []
-        if readonly:
-            if persist:
-                raise RuntimeError("READONLY requests cannot write (duh!) and should not "
-                                   "have PERSISTing values")
-            self.commands = ['AI.DAGRUN_RO']
-        else:
-            self.commands = ['AI.DAGRUN']
-        if load:
-            if not isinstance(load, (list, tuple)):
-                self.commands += ["LOAD", 1, load]
-            else:
-                self.commands += ["LOAD", len(load), *load]
-        if persist:
-            if not isinstance(persist, (list, tuple)):
-                self.commands += ["PERSIST", 1, persist, '|>']
-            else:
-                self.commands += ["PERSIST", len(persist), *persist, '|>']
-        elif load:
-            self.commands.append('|>')
-        self.executor = executor
-
-    def tensorset(self,
-                  key: AnyStr,
-                  tensor: Union[np.ndarray, list, tuple],
-                  shape: Sequence[int] = None,
-                  dtype: str = None) -> Any:
-        args = builder.tensorset(key, tensor, shape, dtype)
-        self.commands.extend(args)
-        self.commands.append("|>")
-        self.result_processors.append(bytes.decode)
-        return self
-
-    def tensorget(self,
-                  key: AnyStr, as_numpy: bool = True,
-                  meta_only: bool = False) -> Any:
-        args = builder.tensorget(key, as_numpy, meta_only)
-        self.commands.extend(args)
-        self.commands.append("|>")
-        self.result_processors.append(partial(utils.tensorget_postprocessor,
-                                              as_numpy=as_numpy,
-                                              meta_only=meta_only))
-        return self
-
-    def modelrun(self,
-                 key: AnyStr,
-                 inputs: Union[AnyStr, List[AnyStr]],
-                 outputs: Union[AnyStr, List[AnyStr]]) -> Any:
-        args = builder.modelrun(key, inputs, outputs)
-        self.commands.extend(args)
-        self.commands.append("|>")
-        self.result_processors.append(bytes.decode)
-        return self
-
-    def run(self):
-        results = self.executor(*self.commands)
-        out = []
-        for res, fn in zip(results, self.result_processors):
-            out.append(fn(res))
-        return out
+processor = Processor()
 
 
 class Client(StrictRedis):
@@ -96,20 +26,54 @@ class Client(StrictRedis):
     debug : bool
         If debug mode is ON, then each command that is sent to the server is
         printed to the terminal
+    enable_postprocess : bool
+        Flag to enable post processing. If enabled, all the bytestring-ed returns
+        are converted to python strings recursively and key value pairs will be converted
+        to dictionaries. Also note that, this flag doesn't work with pipeline() function
+        since pipeline function could have native redis commands (along with RedisAI
+        commands)
 
     Example
     -------
     >>> from redisai import Client
     >>> con = Client(host='localhost', port=6379)
     """
-    def __init__(self, debug=False, *args, **kwargs):
+
+    REDISAI_COMMANDS_RESPONSE_CALLBACKS = {}
+
+    def __init__(self, debug=False, enable_postprocess=True, *args, **kwargs):
         super().__init__(*args, **kwargs)
         if debug:
             self.execute_command = enable_debug(super().execute_command)
+        self.enable_postprocess = enable_postprocess
 
-    def dag(self, load: Sequence = None, persist: Sequence = None,
-            readonly: bool = False) -> Dag:
-        """ It returns a DAG object on which other DAG-allowed operations can be called. For
+    def pipeline(self, transaction: bool = True, shard_hint: bool = None) -> "Pipeline":
+        """
+        It follows the same pipeline implementation of native redis client but enables it
+        to access redisai operation as well. This function is experimental in the
+        current release.
+
+        Example
+        -------
+        >>> pipe = con.pipeline(transaction=False)
+        >>> pipe = pipe.set('nativeKey', 1)
+        >>> pipe = pipe.tensorset('redisaiKey', np.array([1, 2]))
+        >>> pipe.execute()
+        [True, b'OK']
+        """
+        return Pipeline(
+            self.enable_postprocess,
+            self.connection_pool,
+            self.response_callbacks,
+            transaction=True,
+            shard_hint=None,
+        )
+
+    def dag(
+        self, load: Sequence = None, persist: Sequence = None, readonly: bool = False
+    ) -> "Dag":
+        """
+        It returns a DAG object on which other DAG-allowed operations can be called. For
         more details about DAG in RedisAI, refer to the RedisAI documentation.
 
         Parameters
@@ -141,7 +105,9 @@ class Client(StrictRedis):
         >>> # You can even chain the operations
         >>> result = dag.tensorset(**akwargs).modelrun(**bkwargs).tensorget(**ckwargs).run()
         """
-        return Dag(load, persist, self.execute_command, readonly)
+        return Dag(
+            load, persist, self.execute_command, readonly, self.enable_postprocess
+        )
 
     def loadbackend(self, identifier: AnyStr, path: AnyStr) -> str:
         """
@@ -167,18 +133,21 @@ class Client(StrictRedis):
         'OK'
         """
         args = builder.loadbackend(identifier, path)
-        return self.execute_command(*args).decode()
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.loadbackend(res)
 
-    def modelset(self,
-                 key: AnyStr,
-                 backend: str,
-                 device: str,
-                 data: ByteString,
-                 batch: int = None,
-                 minbatch: int = None,
-                 tag: AnyStr = None,
-                 inputs: Union[AnyStr, List[AnyStr]] = None,
-                 outputs: Union[AnyStr, List[AnyStr]] = None) -> str:
+    def modelset(
+        self,
+        key: AnyStr,
+        backend: str,
+        device: str,
+        data: ByteString,
+        batch: int = None,
+        minbatch: int = None,
+        tag: AnyStr = None,
+        inputs: Union[AnyStr, List[AnyStr]] = None,
+        outputs: Union[AnyStr, List[AnyStr]] = None,
+    ) -> str:
         """
         Set the model on provided key.
 
@@ -223,9 +192,11 @@ class Client(StrictRedis):
         ...              inputs=['a', 'b'], outputs=['mul'], tag='v1.0')
         'OK'
         """
-        args = builder.modelset(key, backend, device, data,
-                                batch, minbatch, tag, inputs, outputs)
-        return self.execute_command(*args).decode()
+        args = builder.modelset(
+            key, backend, device, data, batch, minbatch, tag, inputs, outputs
+        )
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.modelset(res)
 
     def modelget(self, key: AnyStr, meta_only=False) -> dict:
         """
@@ -250,8 +221,8 @@ class Client(StrictRedis):
         {'backend': 'TF', 'device': 'cpu', 'tag': 'v1.0'}
         """
         args = builder.modelget(key, meta_only)
-        rv = self.execute_command(*args)
-        return utils.list2dict(rv)
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.modelget(res)
 
     def modeldel(self, key: AnyStr) -> str:
         """
@@ -273,12 +244,15 @@ class Client(StrictRedis):
         'OK'
         """
         args = builder.modeldel(key)
-        return self.execute_command(*args).decode()
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.modeldel(res)
 
-    def modelrun(self,
-                 key: AnyStr,
-                 inputs: Union[AnyStr, List[AnyStr]],
-                 outputs: Union[AnyStr, List[AnyStr]]) -> str:
+    def modelrun(
+        self,
+        key: AnyStr,
+        inputs: Union[AnyStr, List[AnyStr]],
+        outputs: Union[AnyStr, List[AnyStr]],
+    ) -> str:
         """
         Run the model using input(s) which are already in the scope and are associated
         to some keys. Modelrun also needs the output key name(s) to store the output
@@ -314,7 +288,8 @@ class Client(StrictRedis):
         'OK'
         """
         args = builder.modelrun(key, inputs, outputs)
-        return self.execute_command(*args).decode()
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.modelrun(res)
 
     def modelscan(self) -> List[List[AnyStr]]:
         """
@@ -332,17 +307,22 @@ class Client(StrictRedis):
         >>> con.modelscan()
         [['pt_model', ''], ['m', 'v1.2']]
         """
-        warnings.warn("Experimental: Model List API is experimental and might change "
-                      "in the future without any notice", UserWarning)
+        warnings.warn(
+            "Experimental: Model List API is experimental and might change "
+            "in the future without any notice",
+            UserWarning,
+        )
         args = builder.modelscan()
-        result = self.execute_command(*args)
-        return utils.recursive_bytetransform(result, lambda x: x.decode())
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.modelscan(res)
 
-    def tensorset(self,
-                  key: AnyStr,
-                  tensor: Union[np.ndarray, list, tuple],
-                  shape: Sequence[int] = None,
-                  dtype: str = None) -> str:
+    def tensorset(
+        self,
+        key: AnyStr,
+        tensor: Union[np.ndarray, list, tuple],
+        shape: Sequence[int] = None,
+        dtype: str = None,
+    ) -> str:
         """
         Set the tensor to a key in RedisAI
 
@@ -371,11 +351,16 @@ class Client(StrictRedis):
         'OK'
         """
         args = builder.tensorset(key, tensor, shape, dtype)
-        return self.execute_command(*args).decode()
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.tensorset(res)
 
-    def tensorget(self,
-                  key: AnyStr, as_numpy: bool = True,
-                  meta_only: bool = False) -> Union[dict, np.ndarray]:
+    def tensorget(
+        self,
+        key: AnyStr,
+        as_numpy: bool = True,
+        as_numpy_mutable: bool = False,
+        meta_only: bool = False,
+    ) -> Union[dict, np.ndarray]:
         """
         Retrieve the value of a tensor from the server. By default it returns the numpy
         array but it can be controlled using the `as_type` and `meta_only` argument.
@@ -388,6 +373,9 @@ class Client(StrictRedis):
             If True, returns a numpy.ndarray. Returns the value as a list and the
             metadata in a dictionary if False. This flag also decides how to fetch
             the value from the RedisAI server, which also has performance implications
+        as_numpy_mutable : bool
+            If True, returns a a mutable numpy.ndarray object by copy the tensor data. Otherwise (as long as_numpy=True)
+            the returned numpy.ndarray will use the original tensor buffer and will be for read-only
         meta_only : bool
             If True, the value is not retrieved, only the shape and the type
 
@@ -407,9 +395,15 @@ class Client(StrictRedis):
         """
         args = builder.tensorget(key, as_numpy, meta_only)
         res = self.execute_command(*args)
-        return utils.tensorget_postprocessor(res, as_numpy, meta_only)
+        return (
+            res
+            if not self.enable_postprocess
+            else processor.tensorget(res, as_numpy, as_numpy_mutable, meta_only)
+        )
 
-    def scriptset(self, key: AnyStr, device: str, script: str, tag: AnyStr = None) -> str:
+    def scriptset(
+        self, key: AnyStr, device: str, script: str, tag: AnyStr = None
+    ) -> str:
         """
         Set the script to RedisAI. Action similar to Modelset. RedisAI uses the TorchScript
         engine to execute the script. So the script should have only TorchScript supported
@@ -450,7 +444,8 @@ class Client(StrictRedis):
         'OK'
         """
         args = builder.scriptset(key, device, script, tag)
-        return self.execute_command(*args).decode()
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.scriptset(res)
 
     def scriptget(self, key: AnyStr, meta_only=False) -> dict:
         """
@@ -474,8 +469,8 @@ class Client(StrictRedis):
         {'device': 'cpu'}
         """
         args = builder.scriptget(key, meta_only)
-        ret = self.execute_command(*args)
-        return utils.list2dict(ret)
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.scriptget(res)
 
     def scriptdel(self, key: AnyStr) -> str:
         """
@@ -497,14 +492,16 @@ class Client(StrictRedis):
         'OK'
         """
         args = builder.scriptdel(key)
-        return self.execute_command(*args).decode()
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.scriptdel(res)
 
-    def scriptrun(self,
-                  key: AnyStr,
-                  function: AnyStr,
-                  inputs: Union[AnyStr, Sequence[AnyStr]],
-                  outputs: Union[AnyStr, Sequence[AnyStr]]
-                  ) -> str:
+    def scriptrun(
+        self,
+        key: AnyStr,
+        function: AnyStr,
+        inputs: Union[AnyStr, Sequence[AnyStr]],
+        outputs: Union[AnyStr, Sequence[AnyStr]],
+    ) -> str:
         """
         Run an already set script. Similar to modelrun
 
@@ -532,8 +529,8 @@ class Client(StrictRedis):
         'OK'
         """
         args = builder.scriptrun(key, function, inputs, outputs)
-        out = self.execute_command(*args)
-        return out.decode()
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.scriptrun(res)
 
     def scriptscan(self) -> List[List[AnyStr]]:
         """
@@ -550,10 +547,14 @@ class Client(StrictRedis):
         >>> con.scriptscan()
         [['ket1', 'v1.0'], ['ket2', '']]
         """
-        warnings.warn("Experimental: Script List API is experimental and might change "
-                      "in the future without any notice", UserWarning)
+        warnings.warn(
+            "Experimental: Script List API is experimental and might change "
+            "in the future without any notice",
+            UserWarning,
+        )
         args = builder.scriptscan()
-        return utils.recursive_bytetransform(self.execute_command(*args), lambda x: x.decode())
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.scriptscan(res)
 
     def infoget(self, key: AnyStr) -> dict:
         """
@@ -581,8 +582,8 @@ class Client(StrictRedis):
         'duration': 0, 'samples': 0, 'calls': 0, 'errors': 0}
         """
         args = builder.infoget(key)
-        ret = self.execute_command(*args)
-        return utils.list2dict(ret)
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.infoget(res)
 
     def inforeset(self, key: AnyStr) -> str:
         """
@@ -604,4 +605,14 @@ class Client(StrictRedis):
         'OK'
         """
         args = builder.inforeset(key)
-        return self.execute_command(*args).decode()
+        res = self.execute_command(*args)
+        return res if not self.enable_postprocess else processor.inforeset(res)
+
+
+def enable_debug(f):
+    @wraps(f)
+    def wrapper(*args):
+        print(*args)
+        return f(*args)
+
+    return wrapper
